from pyfactcast.client.config import get_client_configuration

from pyfactcast.client.sync import FactStore, get_synchronous_grpc_client


def get_sync_eventstore(profile: str = "default") -> FactStore:
    return FactStore(
        get_synchronous_grpc_client(get_client_configuration(profile=profile))
    )



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

import pyfactcast.grpc.generated.FactStore_pb2 as FactStore__pb2


class RemoteFactStoreStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.subscribe = channel.unary_stream(
            "/org.factcast.grpc.api.gen.RemoteFactStore/subscribe",
            request_serializer=FactStore__pb2.MSG_SubscriptionRequest.SerializeToString,
            response_deserializer=FactStore__pb2.MSG_Notification.FromString,
        )
        self.fetchById = channel.unary_unary(
            "/org.factcast.grpc.api.gen.RemoteFactStore/fetchById",
            request_serializer=FactStore__pb2.MSG_UUID.SerializeToString,
            response_deserializer=FactStore__pb2.MSG_OptionalFact.FromString,
        )
        self.publish = channel.unary_unary(
            "/org.factcast.grpc.api.gen.RemoteFactStore/publish",
            request_serializer=FactStore__pb2.MSG_Facts.SerializeToString,
            response_deserializer=FactStore__pb2.MSG_Empty.FromString,
        )
        self.serialOf = channel.unary_unary(
            "/org.factcast.grpc.api.gen.RemoteFactStore/serialOf",
            request_serializer=FactStore__pb2.MSG_UUID.SerializeToString,
            response_deserializer=FactStore__pb2.MSG_OptionalSerial.FromString,
        )
        self.handshake = channel.unary_unary(
            "/org.factcast.grpc.api.gen.RemoteFactStore/handshake",
            request_serializer=FactStore__pb2.MSG_Empty.SerializeToString,
            response_deserializer=FactStore__pb2.MSG_ServerConfig.FromString,
        )
        self.enumerateNamespaces = channel.unary_unary(
            "/org.factcast.grpc.api.gen.RemoteFactStore/enumerateNamespaces",
            request_serializer=FactStore__pb2.MSG_Empty.SerializeToString,
            response_deserializer=FactStore__pb2.MSG_StringSet.FromString,
        )
        self.enumerateTypes = channel.unary_unary(
            "/org.factcast.grpc.api.gen.RemoteFactStore/enumerateTypes",
            request_serializer=FactStore__pb2.MSG_String.SerializeToString,
            response_deserializer=FactStore__pb2.MSG_StringSet.FromString,
        )
        self.publishConditional = channel.unary_unary(
            "/org.factcast.grpc.api.gen.RemoteFactStore/publishConditional",
            request_serializer=FactStore__pb2.MSG_ConditionalPublishRequest.SerializeToString,
            response_deserializer=FactStore__pb2.MSG_ConditionalPublishResult.FromString,
        )
        self.invalidate = channel.unary_unary(
            "/org.factcast.grpc.api.gen.RemoteFactStore/invalidate",
            request_serializer=FactStore__pb2.MSG_UUID.SerializeToString,
            response_deserializer=FactStore__pb2.MSG_Empty.FromString,
        )
        self.stateFor = channel.unary_unary(
            "/org.factcast.grpc.api.gen.RemoteFactStore/stateFor",
            request_serializer=FactStore__pb2.MSG_StateForRequest.SerializeToString,
            response_deserializer=FactStore__pb2.MSG_UUID.FromString,
        )
        self.currentTime = channel.unary_unary(
            "/org.factcast.grpc.api.gen.RemoteFactStore/currentTime",
            request_serializer=FactStore__pb2.MSG_Empty.SerializeToString,
            response_deserializer=FactStore__pb2.MSG_CurrentDatabaseTime.FromString,
        )
        self.fetchByIdAndVersion = channel.unary_unary(
            "/org.factcast.grpc.api.gen.RemoteFactStore/fetchByIdAndVersion",
            request_serializer=FactStore__pb2.MSG_UUID_AND_VERSION.SerializeToString,
            response_deserializer=FactStore__pb2.MSG_OptionalFact.FromString,
        )
        self.clearSnapshot = channel.unary_unary(
            "/org.factcast.grpc.api.gen.RemoteFactStore/clearSnapshot",
            request_serializer=FactStore__pb2.MSG_SnapshotId.SerializeToString,
            response_deserializer=FactStore__pb2.MSG_Empty.FromString,
        )
        self.setSnapshot = channel.unary_unary(
            "/org.factcast.grpc.api.gen.RemoteFactStore/setSnapshot",
            request_serializer=FactStore__pb2.MSG_Snapshot.SerializeToString,
            response_deserializer=FactStore__pb2.MSG_Empty.FromString,
        )
        self.getSnapshot = channel.unary_unary(
            "/org.factcast.grpc.api.gen.RemoteFactStore/getSnapshot",
            request_serializer=FactStore__pb2.MSG_SnapshotId.SerializeToString,
            response_deserializer=FactStore__pb2.MSG_OptionalSnapshot.FromString,
        )
        self.stateForSpecsJson = channel.unary_unary(
            "/org.factcast.grpc.api.gen.RemoteFactStore/stateForSpecsJson",
            request_serializer=FactStore__pb2.MSG_FactSpecsJson.SerializeToString,
            response_deserializer=FactStore__pb2.MSG_UUID.FromString,
        )


class RemoteFactStoreServicer(object):
    """Missing associated documentation comment in .proto file."""

    def subscribe(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def fetchById(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def publish(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def serialOf(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def handshake(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def enumerateNamespaces(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def enumerateTypes(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def publishConditional(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def invalidate(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def stateFor(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def currentTime(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def fetchByIdAndVersion(self, request, context):
        """1.2"""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def clearSnapshot(self, request, context):
        """1.3"""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def setSnapshot(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def getSnapshot(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def stateForSpecsJson(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_RemoteFactStoreServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "subscribe": grpc.unary_stream_rpc_method_handler(
            servicer.subscribe,
            request_deserializer=FactStore__pb2.MSG_SubscriptionRequest.FromString,
            response_serializer=FactStore__pb2.MSG_Notification.SerializeToString,
        ),
        "fetchById": grpc.unary_unary_rpc_method_handler(
            servicer.fetchById,
            request_deserializer=FactStore__pb2.MSG_UUID.FromString,
            response_serializer=FactStore__pb2.MSG_OptionalFact.SerializeToString,
        ),
        "publish": grpc.unary_unary_rpc_method_handler(
            servicer.publish,
            request_deserializer=FactStore__pb2.MSG_Facts.FromString,
            response_serializer=FactStore__pb2.MSG_Empty.SerializeToString,
        ),
        "serialOf": grpc.unary_unary_rpc_method_handler(
            servicer.serialOf,
            request_deserializer=FactStore__pb2.MSG_UUID.FromString,
            response_serializer=FactStore__pb2.MSG_OptionalSerial.SerializeToString,
        ),
        "handshake": grpc.unary_unary_rpc_method_handler(
            servicer.handshake,
            request_deserializer=FactStore__pb2.MSG_Empty.FromString,
            response_serializer=FactStore__pb2.MSG_ServerConfig.SerializeToString,
        ),
        "enumerateNamespaces": grpc.unary_unary_rpc_method_handler(
            servicer.enumerateNamespaces,
            request_deserializer=FactStore__pb2.MSG_Empty.FromString,
            response_serializer=FactStore__pb2.MSG_StringSet.SerializeToString,
        ),
        "enumerateTypes": grpc.unary_unary_rpc_method_handler(
            servicer.enumerateTypes,
            request_deserializer=FactStore__pb2.MSG_String.FromString,
            response_serializer=FactStore__pb2.MSG_StringSet.SerializeToString,
        ),
        "publishConditional": grpc.unary_unary_rpc_method_handler(
            servicer.publishConditional,
            request_deserializer=FactStore__pb2.MSG_ConditionalPublishRequest.FromString,
            response_serializer=FactStore__pb2.MSG_ConditionalPublishResult.SerializeToString,
        ),
        "invalidate": grpc.unary_unary_rpc_method_handler(
            servicer.invalidate,
            request_deserializer=FactStore__pb2.MSG_UUID.FromString,
            response_serializer=FactStore__pb2.MSG_Empty.SerializeToString,
        ),
        "stateFor": grpc.unary_unary_rpc_method_handler(
            servicer.stateFor,
            request_deserializer=FactStore__pb2.MSG_StateForRequest.FromString,
            response_serializer=FactStore__pb2.MSG_UUID.SerializeToString,
        ),
        "currentTime": grpc.unary_unary_rpc_method_handler(
            servicer.currentTime,
            request_deserializer=FactStore__pb2.MSG_Empty.FromString,
            response_serializer=FactStore__pb2.MSG_CurrentDatabaseTime.SerializeToString,
        ),
        "fetchByIdAndVersion": grpc.unary_unary_rpc_method_handler(
            servicer.fetchByIdAndVersion,
            request_deserializer=FactStore__pb2.MSG_UUID_AND_VERSION.FromString,
            response_serializer=FactStore__pb2.MSG_OptionalFact.SerializeToString,
        ),
        "clearSnapshot": grpc.unary_unary_rpc_method_handler(
            servicer.clearSnapshot,
            request_deserializer=FactStore__pb2.MSG_SnapshotId.FromString,
            response_serializer=FactStore__pb2.MSG_Empty.SerializeToString,
        ),
        "setSnapshot": grpc.unary_unary_rpc_method_handler(
            servicer.setSnapshot,
            request_deserializer=FactStore__pb2.MSG_Snapshot.FromString,
            response_serializer=FactStore__pb2.MSG_Empty.SerializeToString,
        ),
        "getSnapshot": grpc.unary_unary_rpc_method_handler(
            servicer.getSnapshot,
            request_deserializer=FactStore__pb2.MSG_SnapshotId.FromString,
            response_serializer=FactStore__pb2.MSG_OptionalSnapshot.SerializeToString,
        ),
        "stateForSpecsJson": grpc.unary_unary_rpc_method_handler(
            servicer.stateForSpecsJson,
            request_deserializer=FactStore__pb2.MSG_FactSpecsJson.FromString,
            response_serializer=FactStore__pb2.MSG_UUID.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        "org.factcast.grpc.api.gen.RemoteFactStore", rpc_method_handlers
    )
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class RemoteFactStore(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def subscribe(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_stream(
            request,
            target,
            "/org.factcast.grpc.api.gen.RemoteFactStore/subscribe",
            FactStore__pb2.MSG_SubscriptionRequest.SerializeToString,
            FactStore__pb2.MSG_Notification.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def fetchById(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/org.factcast.grpc.api.gen.RemoteFactStore/fetchById",
            FactStore__pb2.MSG_UUID.SerializeToString,
            FactStore__pb2.MSG_OptionalFact.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def publish(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/org.factcast.grpc.api.gen.RemoteFactStore/publish",
            FactStore__pb2.MSG_Facts.SerializeToString,
            FactStore__pb2.MSG_Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def serialOf(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/org.factcast.grpc.api.gen.RemoteFactStore/serialOf",
            FactStore__pb2.MSG_UUID.SerializeToString,
            FactStore__pb2.MSG_OptionalSerial.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def handshake(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/org.factcast.grpc.api.gen.RemoteFactStore/handshake",
            FactStore__pb2.MSG_Empty.SerializeToString,
            FactStore__pb2.MSG_ServerConfig.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def enumerateNamespaces(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/org.factcast.grpc.api.gen.RemoteFactStore/enumerateNamespaces",
            FactStore__pb2.MSG_Empty.SerializeToString,
            FactStore__pb2.MSG_StringSet.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def enumerateTypes(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/org.factcast.grpc.api.gen.RemoteFactStore/enumerateTypes",
            FactStore__pb2.MSG_String.SerializeToString,
            FactStore__pb2.MSG_StringSet.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def publishConditional(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/org.factcast.grpc.api.gen.RemoteFactStore/publishConditional",
            FactStore__pb2.MSG_ConditionalPublishRequest.SerializeToString,
            FactStore__pb2.MSG_ConditionalPublishResult.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def invalidate(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/org.factcast.grpc.api.gen.RemoteFactStore/invalidate",
            FactStore__pb2.MSG_UUID.SerializeToString,
            FactStore__pb2.MSG_Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def stateFor(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/org.factcast.grpc.api.gen.RemoteFactStore/stateFor",
            FactStore__pb2.MSG_StateForRequest.SerializeToString,
            FactStore__pb2.MSG_UUID.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def currentTime(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/org.factcast.grpc.api.gen.RemoteFactStore/currentTime",
            FactStore__pb2.MSG_Empty.SerializeToString,
            FactStore__pb2.MSG_CurrentDatabaseTime.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def fetchByIdAndVersion(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/org.factcast.grpc.api.gen.RemoteFactStore/fetchByIdAndVersion",
            FactStore__pb2.MSG_UUID_AND_VERSION.SerializeToString,
            FactStore__pb2.MSG_OptionalFact.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def clearSnapshot(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/org.factcast.grpc.api.gen.RemoteFactStore/clearSnapshot",
            FactStore__pb2.MSG_SnapshotId.SerializeToString,
            FactStore__pb2.MSG_Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def setSnapshot(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/org.factcast.grpc.api.gen.RemoteFactStore/setSnapshot",
            FactStore__pb2.MSG_Snapshot.SerializeToString,
            FactStore__pb2.MSG_Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def getSnapshot(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/org.factcast.grpc.api.gen.RemoteFactStore/getSnapshot",
            FactStore__pb2.MSG_SnapshotId.SerializeToString,
            FactStore__pb2.MSG_OptionalSnapshot.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def stateForSpecsJson(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/org.factcast.grpc.api.gen.RemoteFactStore/stateForSpecsJson",
            FactStore__pb2.MSG_FactSpecsJson.SerializeToString,
            FactStore__pb2.MSG_UUID.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )



from itertools import zip_longest
import re
from collections import defaultdict
from uuid import UUID

from pyfactcast.app.business.entities import CollectSpec
from pyfactcast.client.entities import Fact, SubscriptionSpec, VersionedType
from typing import Dict, Iterable, List
from pyfactcast.client.sync import FactStore


def _filter_types(*, fs: FactStore, namespace: str, it: Iterable[str]) -> Iterable[str]:
    if not it:
        return it

    types_for_namespace = fs.enumerate_types(namespace=namespace)
    res = set()

    # TODO: Make the regex stuff beautiful
    regexes = [re.compile(elem) for elem in it]
    for regexp in regexes:
        res.update(list(filter(regexp.fullmatch, types_for_namespace)))

    return res


def subscribe(
    namespace: str,
    follow: bool,
    from_now: bool = False,
    after_fact: UUID = None,
    fact_types: List[str] = [],  # noqa: B006 this is not written to
    type_versions: List[int] = [],  # noqa: B006 this is not written to
    *,
    fact_store: FactStore = FactStore(),
) -> Iterable[Fact]:

    if len(type_versions) > len(fact_types):
        raise ValueError(
            (
                "More versions than event types given."
                "Please provide the same number of types and versions or more types than versions."
            )
        )

    if type_versions:
        lookup_types = [
            SubscriptionSpec(
                ns=namespace, type=VersionedType(name=type_name, version=version)
            )
            for type_name, version in zip_longest(
                fact_types, type_versions, fillvalue=0
            )
        ]

    with fact_store as fs:

        if not type_versions:
            resolved_types = _filter_types(fs=fs, namespace=namespace, it=fact_types)
            lookup_types = [
                SubscriptionSpec(
                    ns=namespace, type=VersionedType(name=type_name, version=version)
                )
                for type_name, version in zip_longest(
                    resolved_types, type_versions, fillvalue=0
                )
            ]

        if not fact_types:
            lookup_types = [SubscriptionSpec(ns=namespace)]

        return fs.subscribe(
            subscription_specs=lookup_types,
            continuous=follow,
            from_now=from_now,
            after_fact=after_fact,
        )


def collect(
    collect_specs: List[CollectSpec], *, fact_store: FactStore = FactStore()
) -> Iterable[Fact]:

    subscription_specs = [
        SubscriptionSpec(
            ns=collect_spec.ns,
            type=VersionedType(name=collect_spec.type, version=collect_spec.version),
        )
        for collect_spec in collect_specs
    ]

    with fact_store as fs:
        return fs.subscribe(subscription_specs=subscription_specs)


def collect_by_namespace(
    collect_specs: List[CollectSpec], *, fact_store: FactStore = FactStore()
) -> Dict[str, Iterable[Fact]]:

    spec_by_namespace = defaultdict(lambda: [])
    for spec in collect_specs:
        spec_by_namespace[spec.ns].append(spec)

    result = {}
    with fact_store as fs:
        for ns, specs in spec_by_namespace.items():
            subscription = collect(
                collect_specs=specs, fact_store=fs
            )  # IMPR: use Async client once available
            result[ns] = subscription

    return result



from __future__ import annotations
import re

from typing import Type, TypeVar, Dict, List, Optional
from uuid import uuid4

from pyfactcast.grpc.generated.FactStore_pb2 import MSG_Notification
from pydantic import BaseModel, Extra, constr, validator

import json

F = TypeVar("F", bound="Fact")


class CatchUp:
    pass


UUID4_REGEX = re.compile(
    r"^[a-fA-F0-9]{8}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{12}$"
)
# Sadly pydantic and mypy do not play nice here.
AggsId: constr = constr(  # type: ignore
    regex=r"^[a-fA-F0-9]{8}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{12}$"
)


class FactHeader(BaseModel):
    class Config:
        extra = Extra.allow
        validate_assignment = True

    ns: str
    type: str
    id: str = ""  # type: ignore
    aggIds: Optional[List[AggsId]] = None  # type: ignore
    meta: Dict[str, str] = {}

    @validator("id", always=True)
    def set_id(cls, id: str) -> str:
        if id and not UUID4_REGEX.match(id):
            raise ValueError("id is not a valid UUID4")

        return id or str(uuid4())


class Fact(BaseModel):
    header: FactHeader
    payload: Dict

    @classmethod
    def from_msg(cls: Type[F], msg: MSG_Notification) -> F:
        return cls(
            header=json.loads(msg.fact.header), payload=json.loads(msg.fact.payload)
        )


class VersionedType(BaseModel):
    name: str
    version: int = 0


class SubscriptionSpec(BaseModel):
    ns: str
    type: Optional[VersionedType] = None



from dataclasses import dataclass
from typing import Dict, Optional, Union, cast
from pathlib import Path

import logging
import json
from os import environ as env


log_level = getattr(
    logging, env.get("FACTCAST_LOG_LEVEL", "WARNING").upper(), 30
)  # Setting a default here is pretty defensive. Anyhow nothing lost by doing it

log = logging.getLogger()
log.setLevel(log_level)

# Set grpc dns resolution to native. C-Ares is
env["GRPC_DNS_RESOLVER"] = "native"


@dataclass
class Credentials:
    username: str
    password: str

    def __post_init__(self) -> None:
        if self.username and not self.password:
            raise ValueError(
                "Username was provided without password. Please provide a non empty password."
            )
        if self.password and not self.username:
            raise ValueError(
                "Password was provided without a username. Please provide a non empty username."
            )
        if not self.username and not self.password:
            raise ValueError(
                "Both username and password are empty. Do not instantiate Credentials like this."
            )


@dataclass
class ClientConfiguration:
    server: Optional[str]
    root_cert_path: Optional[str] = None
    ssl_target_override: Optional[str] = None
    credentials: Optional[Credentials] = None
    insecure: bool = False
    default: bool = False

    def __post_init__(self) -> None:
        if not self.server:
            raise ValueError("Server connection string missing.")


def get_client_configuration(profile: str = "default") -> ClientConfiguration:
    log.info("Getting client configuration")

    file_based_config = _get_client_config_from_file()
    profile_from_file: Optional[Dict[str, Union[str, bool]]] = {}
    if file_based_config:
        profiles = file_based_config["profiles"]
        profile_from_file = profiles.get(profile)

        if profile != "default" and not profile_from_file:
            raise ValueError(f"The specified profile ({profile}) does not exist.")

    env_config = _get_client_config_from_env()

    combined_config = {**profile_from_file, **env_config}  # type: ignore # Upgrade 3.9 | notation

    username = cast(str, combined_config.get("grpc_user", ""))
    password = cast(str, combined_config.get("grpc_password", ""))
    credentials = None

    if username or password:
        log.debug("Setting up credentials")
        credentials = Credentials(username, password)

    return ClientConfiguration(
        server=str(combined_config.get("grpc_server", "")),
        credentials=credentials,
        root_cert_path=combined_config.get("grpc_root_cert_path"),  # type: ignore
        ssl_target_override=combined_config.get("grpc_cn_overwrite"),  # type: ignore
        insecure=bool(combined_config.get("grpc_insecure", False)),
    )


def _get_client_config_from_file(
    config_file_path: Optional[Path] = None,
) -> Optional[Dict[str, Dict[str, Dict[str, Union[str, bool]]]]]:

    if not config_file_path:
        config_file_path = Path.home().joinpath(".pyfactcast").absolute()

    if not config_file_path.exists():
        log.info("No configuration file found.")
        return None

    with open(config_file_path, "r") as f:
        config: Dict[str, Dict[str, Dict[str, Union[str, bool]]]] = json.load(f)

    return config


def _get_client_config_from_env() -> Dict[str, Optional[str]]:
    log.info("Getting client configuration from environment")
    result = {}

    # Upgrade 3.8 Walrus
    if env.get("GRPC_SERVER"):
        result["grpc_server"] = env.get("GRPC_SERVER")
    if env.get("GRPC_USER"):
        result["grpc_user"] = env.get("GRPC_USER")
    if env.get("GRPC_PASSWORD"):
        result["grpc_password"] = env.get("GRPC_PASSWORD")
    if env.get("GRPC_ROOT_CERT_PATH"):
        result["grpc_root_cert_path"] = env.get("GRPC_ROOT_CERT_PATH")
    if env.get("GRPC_CN_OVERWRITE"):
        result["grpc_cn_overwrite"] = env.get("GRPC_CN_OVERWRITE")
    if env.get("GRPC_INSECURE"):
        result["grpc_insecure"] = env.get("GRPC_INSECURE")

    return result



from typing import List
import typer
from rich.console import Console
import logging
from rich.logging import RichHandler

from .enumerate import app as enum_app
from .streams import app as sub_app
from .fact import app as fac_app

from ..business.cli import get_sync_eventstore


console = Console()
app = typer.Typer()

FORMAT = "%(message)s"


@app.callback()
def main(
    ctx: typer.Context,
    verbose: List[bool] = typer.Option(
        False,
        "--verbose",
        "-v",
        help="Output more details on what is happening. Use multiple times for more details.",
    ),
    profile: str = typer.Option(
        "default",
        "--profile",
        "-p",
        help="The profile to use for the command that is about to run.",
    ),
) -> None:
    level = "WARNING"
    if len(verbose) == 1:
        level = "INFO"
    if len(verbose) >= 2:
        level = "DEBUG"

    logging.basicConfig(
        level=level, format=FORMAT, datefmt="[%X]", handlers=[RichHandler()]
    )

    ctx.obj = {"fact_store": get_sync_eventstore(profile=profile)}


app.add_typer(enum_app, name="enumerate")
app.add_typer(sub_app, name="streams")
app.add_typer(fac_app, name="fact")


typer_click_object = typer.main.get_command(app)

if __name__ == "__main__":
    app()



from pathlib import Path
from pyfactcast.client.entities import Fact
from typing import List, Optional
from uuid import UUID
from pyfactcast.client.sync import FactStore
from pydantic import parse_file_as


def serial_of(fact_id: UUID, *, fact_store: FactStore = FactStore()) -> Optional[str]:
    with fact_store as fs:
        return fs.serial_of(fact_id=fact_id)


def publish(fact_file: Path, *, fact_store: FactStore = FactStore()) -> None:
    facts = parse_file_as(List[Fact], fact_file)
    for fact in facts:
        fact_store.publish(fact=fact)  # IMPR: Async or as list



import json
import logging
import grpc
import struct
from types import TracebackType
from typing import Dict, Iterable, List, Optional, Type, Union
from json import dumps
from uuid import UUID


from pyfactcast.grpc.generated.FactStore_pb2_grpc import RemoteFactStoreStub
from pyfactcast.grpc.generated.FactStore_pb2 import (
    MSG_Empty,
    MSG_Fact,
    MSG_Facts,
    MSG_String,
    MSG_SubscriptionRequest,
    MSG_Notification,
    MSG_UUID,
)

from pyfactcast.client.auth.basic import BasicAuth
from pyfactcast.client.config import (
    ClientConfiguration,
    get_client_configuration,
    log_level,
)
from pyfactcast.client.entities import Fact, SubscriptionSpec

log = logging.getLogger()
log.setLevel(log_level)


def get_synchronous_grpc_client(
    client_configuration: Optional[ClientConfiguration] = None,
) -> RemoteFactStoreStub:
    log.info("Getting sync client")

    if not client_configuration:
        client_configuration = get_client_configuration()

    if client_configuration.insecure:
        log.warning("Using insecure connection. Only do this for testing!")
        return RemoteFactStoreStub(grpc.insecure_channel(client_configuration.server))

    options = None
    if client_configuration.ssl_target_override:
        log.debug("Setting SSL name override")
        options = (
            (
                "grpc.ssl_target_name_override",
                client_configuration.ssl_target_override,
            ),
        )

    root_cert = None
    if client_configuration.root_cert_path:
        log.debug("Setting SSL certificate path")
        with open(client_configuration.root_cert_path) as f:
            root_cert = f.read().encode("UTF-8")

    channel_credentials = grpc.ssl_channel_credentials(root_certificates=root_cert)

    call_credentials = _construct_call_credentials(client_configuration)

    if call_credentials:
        log.debug("Generating composite credentials")
        grpc_credentials = grpc.composite_channel_credentials(
            channel_credentials, call_credentials
        )
    else:
        grpc_credentials = channel_credentials

    log.debug("Creating channel")
    channel = grpc.secure_channel(
        target=client_configuration.server,
        credentials=grpc_credentials,
        options=options,
    )

    log.debug("Returning stub")
    return RemoteFactStoreStub(channel)


def _construct_call_credentials(
    client_configuration: ClientConfiguration,
) -> Optional[grpc.CallCredentials]:

    if client_configuration.credentials:  # Upgrade 3.8 Walrus
        call_credentials = grpc.metadata_call_credentials(
            BasicAuth(
                client_configuration.credentials.username,
                client_configuration.credentials.password,
            )
        )
        return call_credentials
    return None


def _fact_filter(msg: MSG_Notification) -> bool:
    if msg.type == msg.Fact:
        return True
    elif msg.type == msg.Catchup:
        log.info("Caught up!")
    elif msg.type == msg.Facts:
        log.error("Multiple facts per Notification not implemented.")
    elif msg.type == msg.Complete:
        log.debug("Received stream complete.")
    else:
        log.debug(f"Received Notification of type: {msg.type}")
    return False


class FactStore:
    """The pythonic way of representing a factstore

    Args:
        client (Optional[RemoteFactStoreStub], optional): A remote client stub that
            points towards the server. It will be provisioned on use if None.
            Defaults to None.
    """

    def __init__(self, client: Optional[RemoteFactStoreStub] = None) -> None:
        """ """
        self._client = client

    def __enter__(self) -> "FactStore":
        log.info("Entering FactStore")
        if not self._client:
            self.client = get_synchronous_grpc_client()
        else:
            self.client = self._client
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Optional[bool]:
        # TODO: Implement proper channel termination
        pass

    def enumerate_namespaces(self) -> List[str]:
        """List all namespaces the current user has access to.

        Returns:
            List[str]: All namespaces available to the current user.
        """
        res: List[str] = self.client.enumerateNamespaces(MSG_Empty()).embeddedString

        return res

    def enumerate_types(self, namespace: str) -> List[str]:
        """Provide a list of all event types of a given namespace.

        Args:
            namespace (str): The namespace to enumerate.

        Returns:
            List[str]: All event types in the given namespace.
        """
        message = MSG_String()
        message.embeddedString = namespace
        res: List[str] = self.client.enumerateTypes(message).embeddedString

        return res

    def subscribe(
        self,
        *,
        subscription_specs: Iterable[SubscriptionSpec],
        continuous: bool = False,
        from_now: bool = False,
        after_fact: Optional[UUID] = None,
    ) -> Iterable[Fact]:
        """Creates a fact stream subscription.

        Args:
            subscription_specs (Iterable[SubscriptionSpec]): Used select namespaces and
                to filter the stream for specific versions and types of events.
            continuous (bool, optional): If True the connection will stay open and stream
                new facts as they are received from the factstore. Defaults to False.
            from_now (bool, optional): If True events will only be retrieved from now going forward
                i.e. you will not get any history. Defaults to False.
            after_fact (Optional[UUID], optional): If given and it exists, you will only get facts published
                after the given fact instead of from the beginning of time onwards. Defaults to None.

        Returns:
            Iterable[Fact]: The resulting fact stream.
        """

        specs: List[Dict[str, Union[str, int]]] = []

        for sub_spec in subscription_specs:
            if sub_spec.type:
                specs.append(
                    {
                        "ns": sub_spec.ns,
                        "type": sub_spec.type.name,
                        "version": sub_spec.type.version,
                    }
                )
            else:
                specs.append({"ns": sub_spec.ns})

        log.info(f"Subscribing to {specs}")

        subscription_request_body = {
            "specs": specs,
            "continuous": continuous,
        }

        if from_now:
            subscription_request_body["ephemeral"] = True
            subscription_request_body["continuous"] = True

        if after_fact:
            subscription_request_body["startingAfter"] = str(after_fact)

        msg = MSG_SubscriptionRequest(json=dumps(subscription_request_body))

        log.debug("Calling client subscribe")
        res = self.client.subscribe(msg)
        log.debug(f"Got Result: {res}")

        return map(Fact.from_msg, filter(_fact_filter, res))

    def serial_of(self, *, fact_id: UUID) -> Optional[str]:
        """Returns the serial id of the fact with the given UUID.

        If the fact does not exist will return None.

        Args:
            fact_id (UUID): The UUID of the fact you need a serial for.

        Returns:
            Optional[str]: The serial of the fact as string or None.
        """
        msb, lsb = struct.unpack(">qq", fact_id.bytes)
        msg = MSG_UUID()
        msg.lsb = lsb
        msg.msb = msb
        res = self.client.serialOf(msg)

        if res.present:
            return str(res.serial)
        return None

    def publish(self, *, fact: Fact, conditional: bool = False) -> None:
        """Publishes a fact.

        Args:
            fact (Fact): The fact to publish
            conditional (bool, optional): Whether or not the publish is conditional.
                Defaults to False.

        Raises:
            NotImplementedError: Conditional publication has not been implemented yet.
        """
        if conditional:
            raise NotImplementedError()
        else:
            fact_msg = MSG_Fact(
                header=fact.header.json(), payload=json.dumps(fact.payload)
            )
            msg = MSG_Facts(fact=[fact_msg])
            self.client.publish(msg)



from grpc import AuthMetadataPlugin

from . import AUTHORIZATION_HEADER, BASIC_AUTH_PREFIX
from base64 import b64encode


class BasicAuth(AuthMetadataPlugin):
    def __init__(self, username: str, password: str) -> None:
        encoded = b64encode(f"{username}:{password}".encode("UTF-8")).decode("UTF-8")
        self.auth_string = f"{BASIC_AUTH_PREFIX} {encoded}"

    def __call__(self, context, callback) -> None:  # noqa: ANN001
        """Implements authentication by passing metadata to a callback.

        Implementations of this method must not block.

        Args:
            context: An AuthMetadataContext providing information on the RPC that
            the plugin is being called to authenticate.
            callback: An AuthMetadataPluginCallback to be invoked either
            synchronously or asynchronously.
        """

        callback(((AUTHORIZATION_HEADER, self.auth_string),), None)



import typer
from rich.console import Console

import pyfactcast.app.business.enumerate as business

app = typer.Typer()
console = Console()


@app.command()
def namespaces(
    ctx: typer.Context,
) -> None:
    """
    List all available namespaces in alphabetical order.
    """
    for namespace in sorted(business.namespaces(fact_store=ctx.obj["fact_store"])):
        console.print(namespace, style="bold green")


@app.command()
def types(
    ctx: typer.Context,
    namespace: str = typer.Argument(..., help="A valid namespace you have access to"),
) -> None:
    """
    List all types in a given namespace in alphabetical order
    """
    for type in sorted(business.types(namespace, fact_store=ctx.obj["fact_store"])):
        console.print(type, style="bold green")



from pathlib import Path
from uuid import UUID
import typer
from rich.console import Console

import pyfactcast.app.business.fact as business

app = typer.Typer()
console = Console()


@app.command()
def serial_of(ctx: typer.Context, fact_id: UUID) -> None:
    """
    Returns the serial for the fact identified by the given UUID
    """
    serial = business.serial_of(fact_id, fact_store=ctx.obj["fact_store"])
    if serial:  # Upgrade 3.8 Walrus
        console.print(serial)
    else:
        console.print(f"No fact with id {fact_id} found.")


@app.command()
def publish(
    ctx: typer.Context,
    fact_file: Path = typer.Argument(
        ...,
        exists=True,
        file_okay=True,
        dir_okay=False,
        readable=True,
        resolve_path=True,
        allow_dash=True,
    ),
) -> None:
    """
    Publish the given facts.
    """
    business.publish(fact_file, fact_store=ctx.obj["fact_store"])



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['pyfactcast',
 'pyfactcast.app',
 'pyfactcast.app.business',
 'pyfactcast.app.ui',
 'pyfactcast.client',
 'pyfactcast.client.auth',
 'pyfactcast.grpc',
 'pyfactcast.grpc.generated']

package_data = \
{'': ['*'], 'pyfactcast.grpc': ['proto/*']}

install_requires = \
['grpcio-tools>=1.37.1,<2.0.0',
 'grpcio>=1.37.1,<2.0.0',
 'pydantic>=1.8.2,<2.0.0',
 'rich>=10.2.0,<11.0.0',
 'typer>=0.3.2,<0.4.0']

extras_require = \
{'docs': ['sphinx<4',
          'sphinx-click>=2.7.1,<3.0.0',
          'sphinx-rtd-theme>=0.5.2,<0.6.0',
          'sphinx-autodoc-typehints>=1.12.0,<2.0.0']}

entry_points = \
{'console_scripts': ['factcast = pyfactcast.app.ui.cli:app']}

setup_kwargs = {
    'name': 'pyfactcast',
    'version': '0.0.9',
    'description': 'A python client library for FactCast',
    'long_description': '# PyFactCast\n\nWelcome to pyfactcast. You can find more extensive documentation over at [readthedocs](https://pyfactcast.readthedocs.io/en/latest/).\n\nThis project arose manly out of frustration with the excessive wait times for a spring boot\nbased CLI like the one offered by the original [FactCast](https://docs.factcast.org/).\nBut on the way grew a bit. It will likely grow even more as I would like to actually use this\ntooling to bring the ability for rapid prototyping and production grade python applications\nto the FactCast community.\n\nContributions are welcome. Just get in touch.\n\n## Quickstart\n\nSimply `pip install pyfactcast` and get going. The cli is available as `factcast` and\nyou can run `factcast --help` to get up to speed on what you can do.\n\n## Development\n\nThis project uses `poetry` for dependency management and `pre-commit` for local checks.\n',
    'author': 'Eduard Thamm',
    'author_email': 'eduard.thamm@thammit.at',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/edthamm/pyfactcast',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from pathlib import Path
from typing import Iterable, List, Optional
from uuid import UUID
from pydantic import parse_file_as
import typer
from rich.console import Console
from rich.table import Table
from datetime import datetime
import json
from pyfactcast.app.business.entities import CollectSpec

import pyfactcast.app.business.streams as business

app = typer.Typer()
console = Console()


def print_iterable_as_json_list(iterable: Iterable) -> None:
    console.print("[")
    try:
        it = iter(iterable)

        last = next(it)  # type: ignore
        console.print(f"{last.json(indent=2)}", soft_wrap=True)

        for elem in it:
            last = elem
            console.print(f",{last.json(indent=2)}", soft_wrap=True)
    except StopIteration:
        return
    finally:
        console.print("]")


def write_iterable_as_json_list(iterable: Iterable, out: Path) -> None:
    with open(out, "w") as f:
        f.write("[")
        try:
            it = iter(iterable)

            last = next(it)  # type: ignore
            f.write(f"{last.json(indent=2)}")

            for elem in it:
                last = elem
                f.write(f",{last.json(indent=2)}")
        except StopIteration:
            return
        finally:
            f.write("]")


@app.command()
def subscribe(
    ctx: typer.Context,
    namespace: str,
    follow: bool = typer.Option(
        False, "--follow", help="Stay connected and listen for new events."
    ),
    from_now: bool = typer.Option(
        False, "--from-now", help="Follow the event stream starting now."
    ),
    after_fact: UUID = typer.Option(
        None, help="Get all events in the stream that happened after a certain ID."
    ),
    json_format: bool = typer.Option(False, "--json", help="Output raw json events."),
    fact_types: List[str] = typer.Option(
        [],
        "--type",
        help="The types of events to get. Can be used multiple times. Supports regexp.",
    ),
    type_versions: List[int] = typer.Option(
        [],
        "--version",
        help="The versions of the types to get. Will be right padded with latest. If used regex for types is not supported.",
    ),
) -> None:
    """
    Subscribe to an eventstream.

    You can select types you want to subscribe to. When doing so you can also use regexes to subscribe to multiple
    event types in short order for example ``... subscribe machinery --type Assembly.*``
    will get you all Assembly events for example AssemblyCreated, AssemblyCompleted, AssemblyDestroyed but nothing related
    to Subassemblies. If you specify no types, all types will be streamed.

    You can also get event types in specific versions.
    You can either specify them intermixed with the types like this
    ``... subscribe machinery --type AssemblyCreated --version 1 --type AssemblyUsed --version 3``
    or you can write them out after each other like this
    ``... subscribe machinery --type AssemblyCreated --type AssemblyUsed --version 1 --version 3``.
    In both cases you will get version 1 of AssemblyCreated and version 3 of AssemblyUsed.
    You can specify fewer versions than types. In this case matching will be left to right by postion in the list.
    Types without a version will give you the latest version. This leads to a slightly strange case where
    ``... subscribe machinery --type AssemblyCreated --type AssemblyUsed --version 2``
    will give you version 2 of AssemblyCreated and the latest version of AssemblyUsed.
    It is recommended to put the types you need the latest version of last and use intermixed notation before that.
    If you want to explicitly specify the latest version, please specify ``--version 0``


    To keep other complexities out of this, if you use versions you will not be able
    to use regexes. This is mostly due to the ambiguities that come along when you specify something like this
    ``... subscribe machinery --type Assembly.* --version 2`` do you want version 2 of everything or version 2 of the first match
    and latest of everything else?

    If you are looking for information on ``tee`` and ``--follow`` please check the
    :ref:`Streams FAQ<faq/streams:Fact Stream FAQ>`.
    """
    subscription = business.subscribe(
        namespace=namespace,
        follow=follow,
        after_fact=after_fact,
        from_now=from_now,
        fact_types=fact_types,
        type_versions=type_versions,
        fact_store=ctx.obj["fact_store"],
    )

    if json_format:
        print_iterable_as_json_list(subscription)
        return

    for elem in subscription:
        table = Table(show_header=False, show_lines=True, header_style="bold magenta")
        table.add_column("Time")
        table.add_column("Type")
        table.add_column("Header")
        table.add_column("Payload")
        table.add_row(
            datetime.fromtimestamp(int(elem.header.meta["_ts"][:10])).isoformat(),  # type: ignore
            elem.header.type,
            elem.header.json(indent=2),
            json.dumps(elem.payload, indent=2),
        )
        console.print(table)


@app.command()
def collect(
    ctx: typer.Context,
    collect_spec: Path,
    out: Optional[Path] = typer.Option(
        None, help="The output file or folder to which the result is to be delivered."
    ),
    by_namespace: bool = typer.Option(
        False,
        "--by-namespace",
        help='If set the result will be output to one file per namespace. Requires "out" option to be a directory.',
    ),
) -> None:
    """
    This command helps in piecing together more complex scenarios. It takes a file containing
    a list of dicts ``[{"ns": "testNamespace", "type": "testType", "version":2}]`` where ``version``
    is optional. The type is not regexp-able. This can be used to gather the base data for projections
    which can assist in debugging.

    The command will collect all events matching the specification into one event stream and either
    print it to screen of output it to a file.
    If you need the eventstreams separated by namespace you MUST pass an output directroy along
    with the ``--by--namespace`` option. The resulting eventstreams will be stored as json lists
    in that directory in files named after the namespaces.
    """
    collect_specs = parse_file_as(List[CollectSpec], collect_spec)
    fs = ctx.obj["fact_store"]

    if by_namespace:
        if not out:
            console.print(
                "--by-namespace is set. Please provide a directory through --out.",
                style="red",
            )
            raise typer.Exit(1)
        if not out.is_dir():
            console.print(
                "--by-namespace is set. --out must be a directory.", style="red"
            )
            raise typer.Exit(1)

        namespaced_results = business.collect_by_namespace(
            collect_specs=collect_specs, fact_store=fs
        )
        for ns, result in namespaced_results.items():
            write_iterable_as_json_list(result, out.joinpath(f"{ns}.json"))

    subscription = business.collect(collect_specs=collect_specs, fact_store=fs)

    if not out:
        print_iterable_as_json_list(subscription)

    if out and not by_namespace:
        write_iterable_as_json_list(subscription, out)
